package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc;

import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;

/**
 * This class represents an event driven process chain without events and or
 * connectors because they are not required for our case.
 * 
 * @author thomas
 * 
 */
public class EPC {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(EPC.class);

	/**
	 * Set of EPCFunction contains all functions of this EPC
	 */
	private Vector<EPCFunction> functions;

	/**
	 * Set of EPCAndConnector contains all and connectors of this EPC
	 */
	private Vector<EPCAndConnector> andConnectors;

	/**
	 * Set of EPCXorConnector contains all xor connectors of this EPC
	 */
	private Vector<EPCXorConnector> xorConnectors;

	private Vector<LooseEnd> looseEnds;

	HashMap<EPCNode, EPCNode> mapping = null;

	/**
	 * Creates a new EPC object
	 */
	public EPC() {

		functions = new Vector<EPCFunction>();

		andConnectors = new Vector<EPCAndConnector>();

		xorConnectors = new Vector<EPCXorConnector>();

		looseEnds = new Vector<LooseEnd>();
	}

	public LooseEnd addLooseEnd(Activity from, Activity to, boolean out) {
		LooseEnd loose = new LooseEnd(this, from, to, out);
		looseEnds.add(loose);

		return loose;
	}

	public void removeAllLoseEnds() {
		for (LooseEnd end : looseEnds) {
			if (!end.getPreset().isEmpty() || !end.getPostset().isEmpty()) {
				LOG.error("The loose end " + end
						+ " is still connected to the EPC!");
				return;
			}
		}
		looseEnds.clear();
	}

	public Vector<LooseEnd> getLooseEnds() {
		return looseEnds;
	}

	/**
	 * Adds a new function with the given name to this EPC.
	 * 
	 * @param label
	 *            label of the function as String
	 * @return EPCFunction object which represents this function
	 */
	public EPCFunction addFunction(String label) {
		EPCFunction function = new EPCFunction(label, this);

		functions.add(function);

		return function;
	}

	/**
	 * Adds a new and connector to this EPC.
	 * 
	 * @return EPCAndConnector object which represents this and connector
	 */
	public EPCAndConnector addAndConnector() {
		EPCAndConnector andConnector = new EPCAndConnector(this);

		andConnectors.add(andConnector);

		return andConnector;
	}

	/**
	 * Adds a new xor connector to this EPC.
	 * 
	 * @return EPCXorConnector object which represents this xor connector
	 */
	public EPCXorConnector addXorConnector() {
		EPCXorConnector xorConnector = new EPCXorConnector(this);

		xorConnectors.add(xorConnector);

		return xorConnector;
	}

	/**
	 * Removes all connectors from the EPC, that have only one incoming edge and
	 * one outgoing edge.
	 */
	public void removeUselessConnectors() {

		Vector<EPCConnector> connectors = findUselessConnectors();

		for (EPCConnector connector : connectors) {
			if (connector.getPostset().size() == 1
					&& connector.getPreset().size() == 1) {
				EPCEdge incomingEdge = connector.getPreset().firstElement();
				EPCEdge outgoingEdge = connector.getPostset().firstElement();

				EPCNode predecessor = incomingEdge.getStart();
				EPCNode successor = outgoingEdge.getEnd();

				predecessor.removeOutgoingEdge(incomingEdge);
				successor.removeIncomingEdge(outgoingEdge);

				predecessor.connectToSuccessor(successor);
			}

			EPCEdge.deleteAllEdges(connector);
		}

		xorConnectors.removeAll(connectors);
		andConnectors.removeAll(connectors);
	}

	/**
	 * Find "useless" connectors. This means a set is returned with all
	 * connectors which have only one incoming and one outgoing edge.
	 * 
	 * @return {@link Vector} of connectors with one outgoing and one incoming
	 *         edge
	 */
	private Vector<EPCConnector> findUselessConnectors() {

		Vector<EPCConnector> connectors = new Vector<EPCConnector>();

		for (EPCConnector connector : andConnectors) {
			if (connector.getPreset().size() == 1
					&& connector.getPostset().size() == 1) {
				connectors.add(connector);
			}
		}

		for (EPCConnector connector : xorConnectors) {
			if (connector.getPreset().size() == 1
					&& connector.getPostset().size() == 1) {
				connectors.add(connector);
			}
		}

		for (EPCConnector connector : andConnectors) {
			if (connector.getPreset().size() == 0
					|| connector.getPostset().size() == 0) {
				connectors.add(connector);
			}
		}

		for (EPCConnector connector : xorConnectors) {
			if (connector.getPreset().size() == 0
					|| connector.getPostset().size() == 0) {
				connectors.add(connector);
			}
		}

		return connectors;
	}

	/**
	 * This method connects the given nodes.
	 */
	public EPCEdge connect(EPCNode start, EPCNode end) {
		return start.connectToSuccessor(end);
	}

	/**
	 * This method disconnects the given nodes.
	 */
	public void disconnect(EPCNode start, EPCNode end) {
		start.disconnectFromSuccessor(end);
	}

	/**
	 * Returns all functions of this EPC.
	 * 
	 * @return functions as Vector of EPCFunction
	 */
	public Vector<EPCFunction> getFunctions() {
		return functions;
	}

	/**
	 * Returns all and connectors of this EPC.
	 * 
	 * @return and connectors as Vector of EPCAndConnectors
	 */
	public Vector<EPCAndConnector> getAndConnectors() {
		return andConnectors;
	}

	public Vector<EPCNode> getAllNodes() {
		Vector<EPCNode> nodes = new Vector<EPCNode>();

		nodes.addAll(andConnectors);
		nodes.addAll(functions);
		nodes.addAll(xorConnectors);

		return nodes;
	}

	/**
	 * Returns all xor connectors of this EPC.
	 * 
	 * @return xor connectors as Vector of EPCXorConnectors
	 */
	public Vector<EPCXorConnector> getXorConnectors() {
		return xorConnectors;
	}

	/**
	 * This method test if a set of nodes can fire concurrent.
	 * 
	 * @param nodes
	 *            set of nodes as {@link Vector} of {@link EPCNode}
	 * @return true if set can fire concurrent
	 */
	public boolean canFire(Vector<EPCNode> nodes) {
		for (EPCNode node : nodes) {
			if (!node.canFire()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * This method sets the marking of all edges to the initial marking. This
	 * means all edges which follow initial events are marked and all other
	 * edges are unmarked.
	 */
	public void markWithInitialMarking() {
		Vector<EPCEdge> edges = getAllEdges();

		for (EPCEdge edge : edges) {
			if (edge.getStart().getPreset().isEmpty()) {
				Vector<EPCFunction> sources = new Vector<EPCFunction>();
				sources.add((EPCFunction) edge.getStart());
				try {
					edge.setMarked(true, sources);
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else {
				try {
					edge.setMarked(false, null);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * This method returns all initial nodes. This are nodes with empty preset.
	 * 
	 * @return set of initial nodes as {@link Vector} of {@link EPCNode}
	 */
	public Vector<EPCNode> getInitialNodes() {
		Vector<EPCNode> initial = new Vector<EPCNode>();

		for (EPCNode node : functions) {
			if (node.getPreset().isEmpty()) {
				initial.add(node);
			}
		}

		for (EPCNode node : xorConnectors) {
			if (node.getPreset().isEmpty()) {
				initial.add(node);
			}
		}

		for (EPCNode node : andConnectors) {
			if (node.getPreset().isEmpty()) {
				initial.add(node);
			}
		}

		return initial;
	}

	public Vector<EPCEdge> getAllEdges() {
		return EPCEdge.getAllEdges(this);
	}

	public boolean isNodeOfThisEpc(EPCNode node) {
		if (andConnectors.contains(node)) {
			return true;
		} else if (functions.contains(node)) {
			return true;
		} else if (looseEnds.contains(node)) {
			return true;
		} else if (xorConnectors.contains(node)) {
			return true;
		}

		return false;
	}

	public void deleteLooseEnd(LooseEnd looseEnd) {
		looseEnds.remove(looseEnd);
	}

	public EPC clone() {
		EPC clone = new EPC();

		clone.mapping = new HashMap<EPCNode, EPCNode>();

		for (EPCNode node : andConnectors) {
			clone.mapping.put(node, clone.addAndConnector());
		}

		for (EPCNode node : xorConnectors) {
			clone.mapping.put(node, clone.addXorConnector());
		}

		for (EPCNode node : functions) {
			clone.mapping.put(node, clone.addFunction(node.getLabel()));
		}

		for (EPCNode node : looseEnds) {
			LooseEnd looseEnd = (LooseEnd) node;
			clone.mapping.put(node, clone.addLooseEnd(looseEnd.getStart(),
					looseEnd.getEnd(), looseEnd.isOutgoingLooseEnd()));
		}

		for (EPCNode node : clone.mapping.keySet()) {
			for (EPCEdge edge : node.getPostset()) {
				EPCEdge edgeClone = clone.connect(clone.mapping.get(node),
						clone.mapping.get(edge.getEnd()));

				try {
					edgeClone.setMarked(edge.isMarked(), edge.getMarkSources());
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		return clone;
	}

	public EPCNode translateToClonedNode(EPCNode node) {
		return mapping.get(node);
	}
}
