package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc;

import java.security.InvalidParameterException;
import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class represents a directed edge in the EPC.
 * 
 * @author thomas
 */
public class EPCEdge {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(EPCEdge.class);

	/**
	 * references to all edges
	 */
	private static HashMap<EPCNode, HashMap<EPCNode, EPCEdge>> edges;

	/**
	 * This method returns the edge form start to end. If no such edge exits it
	 * is created.
	 * 
	 * @param start
	 *            start node
	 * @param end
	 *            end node
	 * @return edge form start to end
	 */
	public static EPCEdge getEdge(EPCNode start, EPCNode end, EPC epc) {
		if (edges == null) {
			edges = new HashMap<EPCNode, HashMap<EPCNode, EPCEdge>>();
		}

		if (!edges.containsKey(start)) {
			edges.put(start, new HashMap<EPCNode, EPCEdge>());
		}

		HashMap<EPCNode, EPCEdge> edgesFromStart = edges.get(start);

		if (!edgesFromStart.containsKey(end)) {
			edgesFromStart.put(end, new EPCEdge(start, end, epc));
		}

		if (!start.getEpc().equals(epc)) {
			throw new InvalidParameterException(
					"The given epc is not the epc of the start node!");
		} else if (!end.getEpc().equals(epc)) {
			throw new InvalidParameterException(
					"The given epc is not the epc of the end node!");
		} else if (!epc.isNodeOfThisEpc(start)) {
			throw new InvalidParameterException(
					"The start node is no node of this epc!");
		} else if (!epc.isNodeOfThisEpc(end)) {
			throw new InvalidParameterException(
					"The end node is no node of this epc!");
		}

		return edgesFromStart.get(end);
	}

	/**
	 * This method search all edges which belongs to the given epc.
	 * 
	 * @param epc
	 *            epc as {@link EPC}
	 * @return {@link Vector} of {@link EPCEdge}s
	 */
	public static Vector<EPCEdge> getAllEdges(EPC epc) {
		Vector<EPCEdge> epcEdges = new Vector<EPCEdge>();

		for (HashMap<EPCNode, EPCEdge> edgeMap : edges.values()) {
			for (EPCEdge edge : edgeMap.values()) {
				if (epc.equals(edge.getEpc())) {
					epcEdges.add(edge);
				}
			}
		}

		return epcEdges;
	}

	/**
	 * This method deletes the edge form start to end from the edge registry.
	 * 
	 * @param start
	 *            start of the edge
	 * @param end
	 *            end of the edge
	 */
	public static EPCEdge deleteEdge(EPCNode start, EPCNode end) {
		if (edges == null) {
			return null;
		}

		if (!edges.containsKey(start)) {
			return null;
		}

		HashMap<EPCNode, EPCEdge> edgesFromStart = edges.get(start);

		EPCEdge edgeToDelete = edgesFromStart.get(end);

		edgesFromStart.remove(end);

		return edgeToDelete;
	}

	public static void deleteAllEdges(EPCNode node) {
		if (edges == null || node == null) {
			return;
		}

		if (edges.containsKey(node)) {
			for (EPCEdge edge : edges.get(node).values()) {
				edge.getEnd().removeIncomingEdge(edge);
				node.removeOutgoingEdge(edge);
			}
		}

		edges.remove(node);

		for (HashMap<EPCNode, EPCEdge> map : edges.values()) {
			if (map.containsKey(node)) {
				EPCEdge edge = map.get(node);
				node.removeIncomingEdge(edge);
				edge.getStart().removeOutgoingEdge(edge);
				map.remove(node);
			}
		}
	}

	/**
	 * node where the edge starts
	 */
	private EPCNode start;

	/**
	 * node where the edge ends
	 */
	private EPCNode end;

	/**
	 * epc where this edge belongs to
	 */
	private EPC epc;

	/**
	 * true if the edge is marked (used for unfolding)
	 */
	private boolean marked;

	private Vector<EPCFunction> markSources;

	/**
	 * Constructs a new edge.
	 * 
	 * @param from
	 *            node where the edge starts
	 * @param to
	 *            node where the edge ends
	 */
	private EPCEdge(EPCNode from, EPCNode to, EPC epc) {
		this.start = from;
		this.end = to;
		this.epc = epc;

		marked = false;
		markSources = new Vector<EPCFunction>();
	}

	/**
	 * This mehtod returns the epc where this edge belongs to.
	 * 
	 * @return epc of this edge
	 */
	public EPC getEpc() {
		return epc;
	}

	/**
	 * This method returns the start node of this edge.
	 * 
	 * @return start of the edge as {@link EPCNode}
	 */
	public EPCNode getStart() {
		return start;
	}

	/**
	 * This method returns the end node of this edge.
	 * 
	 * @return end of the edge as {@link EPCNode}
	 */
	public EPCNode getEnd() {
		return end;
	}

	/**
	 * This method returns true if the edge is marked.
	 * 
	 * @return true if marked
	 */
	public boolean isMarked() {
		return marked;
	}

	/**
	 * This method sets the marking of the edge.
	 * 
	 * @param marked
	 *            true to set marked, false else
	 * @throws Exception
	 */
	public void setMarked(boolean marked, Vector<EPCFunction> markSources)
			throws Exception {
		if (marked) {
			if (markSources == null || markSources.isEmpty()) {
				throw new Exception(
						"If an edge is marked there have to be sources!");
			}

			LOG.debug("Marking edge " + this + " with marks from "
					+ markSources);
		} else {
			LOG.debug("Unmarking edge " + this + ". Edge had marks "
					+ markSources);
		}

		this.marked = marked;
		this.markSources = markSources;
	}

	public Vector<EPCFunction> getMarkSources() {
		return markSources;
	}

	/**
	 * Tests if the other object equals this object. For a edge this means the
	 * start and end node are equal.
	 * 
	 * @return true if both edges are equal
	 */
	public boolean equals(Object other) {
		if (other instanceof EPCEdge) {
			EPCEdge otherEdge = (EPCEdge) other;

			if (otherEdge.start.equals(start) && otherEdge.end.equals(end)) {
				return true;
			}
		}

		return false;
	}

	@Override
	public String toString() {
		return "edge " + " (" + hashCode() + ")";
	}
}
