package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc;

import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Base of all EPC nodes.
 * 
 * @author thomas
 */
public abstract class EPCNode {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(EPCNode.class);

	/**
	 * incomming edges of this connector
	 */
	private Vector<EPCEdge> incoming;

	/**
	 * outgoing edges of this connector
	 */
	private Vector<EPCEdge> outgoing;

	/**
	 * EPC where this node belongs to
	 */
	private EPC epc;

	/**
	 * constructs a new EPC node.
	 */
	public EPCNode(EPC epc) {
		this.epc = epc;

		incoming = new Vector<EPCEdge>();

		outgoing = new Vector<EPCEdge>();
	}

	/**
	 * This method adds an edge as incoming edge of this node.
	 * 
	 * @param edge
	 *            edge to add
	 */
	public void addIncomingEdge(EPCEdge edge) {
		if (!edge.getEnd().equals(this)) {
			LOG.error("Edge doesn't end at this node!");
			return;
		}

		incoming.add(edge);
	}

	/**
	 * This method removes an edge as incoming edge of this node.
	 * 
	 * @param edge
	 *            edge to remove
	 */
	public void removeIncomingEdge(EPCEdge edge) {
		EPCEdge.deleteEdge(edge.getStart(), this);
		incoming.remove(edge);
	}

	/**
	 * This method adds an edge as outgoing edge of this node.
	 * 
	 * @param edge
	 *            edge to add
	 */
	public void addOutgoingEdge(EPCEdge edge) {
		if (!edge.getStart().equals(this)) {
			LOG.error("Edge doesn't start at this node!");
			return;
		}

		outgoing.add(edge);
	}

	/**
	 * This method removes an edge as outgoing edge of this node.
	 * 
	 * @param edge
	 *            edge to remove
	 */
	public void removeOutgoingEdge(EPCEdge edge) {
		EPCEdge.deleteEdge(this, edge.getEnd());
		outgoing.remove(edge);
	}

	/**
	 * This method returns the incoming edges of this node.
	 * 
	 * @return {@link Vector} of incoming edges as {@link EPCEdge}
	 */
	public Vector<EPCEdge> getPreset() {
		return incoming;
	}

	/**
	 * This method returns the outgoing edges of this node.
	 * 
	 * @return {@link Vector} of outgoing edges as {@link EPCEdge}
	 */
	public Vector<EPCEdge> getPostset() {
		return outgoing;
	}

	/**
	 * This method connects this node to the given predecessor.
	 * 
	 * @param predecessor
	 *            predecessor as {@link EPCNode}
	 */
	public void connectToPredecessor(EPCNode predecessor) {
		EPCEdge edge = EPCEdge.getEdge(predecessor, this, epc);
		addIncomingEdge(edge);
		predecessor.addOutgoingEdge(edge);
	}

	/**
	 * This method connects this node to the given successor.
	 * 
	 * @param successor
	 *            successor as {@link EPCNode}
	 */
	public EPCEdge connectToSuccessor(EPCNode successor) {
		EPCEdge edge = EPCEdge.getEdge(this, successor, epc);
		addOutgoingEdge(edge);
		successor.addIncomingEdge(edge);
		return edge;
	}

	/**
	 * This method disconnects this node form the given predecessor.
	 * 
	 * @param predecessor
	 *            predecessor as {@link EPCNode}
	 */
	public void disconnectFromPredecessor(EPCNode predecessor) {
		EPCEdge edge = EPCEdge.deleteEdge(predecessor, this);
		if (edge != null) {
			removeIncomingEdge(edge);
			predecessor.removeOutgoingEdge(edge);
		}
	}

	/**
	 * This method disconnects this node from the given successor.
	 * 
	 * @param successor
	 *            successor as {@link EPCNode}
	 */
	public void disconnectFromSuccessor(EPCNode successor) {
		EPCEdge edge = EPCEdge.deleteEdge(this, successor);
		if (edge != null) {
			removeOutgoingEdge(edge);
			successor.removeIncomingEdge(edge);
		}
	}

	/**
	 * This method returns true if the node can fire.
	 * 
	 * @return true if node can fire
	 */
	public abstract boolean canFire();

	/**
	 * This method fires the node. The parameter choice is used if the firing
	 * rule needs more information.
	 * 
	 * @param choice
	 *            choice is used to select the outgoing edge
	 */
	public abstract void fire(EPCNode choice) throws Exception;

	public abstract String getLabel();

	/**
	 * This method returns the epc where this node belongs to.
	 * 
	 * @return eps as {@link EPC}
	 */
	public EPC getEpc() {
		return epc;
	}
}
