package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.buildingblocks;

import java.rmi.UnexpectedException;
import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCAndConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCFunction;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCNode;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCXorConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.LooseEnd;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;

public class BuildingBlock extends EPC {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory
			.getLogger(BuildingBlock.class);

	/**
	 * activity which is represented by this building block
	 */
	private Activity activity;

	/**
	 * function which represents the action of this block
	 */
	private EPCFunction function;

	/**
	 * split different presets
	 */
	private EPCXorConnector preXor;

	/**
	 * split different postsets
	 */
	private EPCXorConnector postXor;

	/**
	 * map for pre activities
	 */
	private HashMap<Activity, BlockConnector> preActivityMap;

	/**
	 * map for post activities
	 */
	private HashMap<Activity, BlockConnector> postActivityMap;

	private Vector<ActivitySet> presets;

	private Vector<ActivitySet> postsets;

	/**
	 * Constructs a new building block for the given activity.
	 * 
	 * @param activity
	 *            activity represented by this block
	 */
	public BuildingBlock(Activity activity) {
		super();

		this.activity = activity;

		preActivityMap = new HashMap<Activity, BlockConnector>();
		postActivityMap = new HashMap<Activity, BlockConnector>();

		presets = new Vector<ActivitySet>();
		postsets = new Vector<ActivitySet>();

		function = addFunction(activity.getLabel());

		preXor = addXorConnector();
		connect(preXor, function);

		postXor = addXorConnector();
		connect(function, postXor);
	}

	/**
	 * This method adds a possible preset to this building block.
	 * 
	 * @param preset
	 *            preset as {@link Vector} of {@link Activity}
	 */
	public void addPreset(ActivitySet preset) {
		if (preset.isEmpty()) {
			return;
		}

		if (presets.contains(preset)) {
			LOG.debug("The preset " + preset + " of " + this.activity
					+ " was already known and not added again.");
			return;
		}

		presets.add(preset);

		EPCAndConnector presetAnd = addAndConnector();
		connect(presetAnd, preXor);

		for (Activity activity : preset) {
			if (!preActivityMap.containsKey(activity)) {
				EPCXorConnector xor = addXorConnector();

				connect(xor, presetAnd);

				BlockConnector connector = addPreConnector(activity);
				connect(connector, xor);

				preActivityMap.put(activity, connector);
			} else {
				BlockConnector connector = preActivityMap.get(activity);
				EPCNode xor = connector.getPostset().firstElement().getEnd();
				connect(xor, presetAnd);
			}
		}
	}

	public void printBlockToErr() {
		System.err.println(activity + ": xor: " + getXorConnectors().size()
				+ ", and: " + getAndConnectors().size());

		EPCNode node = function;
		String out = "Post: " + node + " -- ";
		while (!node.getPostset().isEmpty()) {
			node = node.getPostset().firstElement().getEnd();
			out += node + " -- ";
		}
		System.err.println(out);

		node = function;
		out = "Pre: " + node + " -- ";
		while (!node.getPreset().isEmpty()) {
			node = node.getPreset().firstElement().getStart();
			out += node + " -- ";
		}
		System.err.println(out);

	}

	/**
	 * This method adds a possible postset to this building block.
	 * 
	 * @param postset
	 *            postset as {@link Vector} of {@link Activity}
	 */
	public void addPostset(ActivitySet postset) {
		if (postset.isEmpty()) {
			return;
		}

		if (postsets.contains(postset)) {
			LOG.debug("The preset " + postset + " of " + this.activity
					+ " was already known and not added again.");
			return;
		}

		postsets.add(postset);

		EPCAndConnector postsetAnd = addAndConnector();
		connect(postXor, postsetAnd);

		for (Activity activity : postset) {
			if (!postActivityMap.containsKey(activity)) {
				EPCXorConnector xor = addXorConnector();
				connect(postsetAnd, xor);

				BlockConnector connector = addPostConnector(activity);
				connect(xor, connector);

				postActivityMap.put(activity, connector);
			} else {
				BlockConnector connector = postActivityMap.get(activity);

				EPCNode xor = connector.getPreset().firstElement().getStart();

				connect(postsetAnd, xor);
			}
		}
	}

	/**
	 * Adds a new incoming {@link BlockConnector} for the {@link Activity} which
	 * is represented by this label.
	 * 
	 * @param label
	 *            label to add connection to
	 */
	public BlockConnector addPreConnector(String label) {
		return addPreConnector(Activity.getActivity(label));
	}

	/**
	 * Adds a new incoming {@link BlockConnector} for the given activity to this
	 * block.
	 * 
	 * @param acivity
	 *            activity to connect to
	 */
	public BlockConnector addPreConnector(Activity acivity) {
		BlockConnector connector = new BlockConnector(this, acivity);
		return addPreConnector(connector);
	}

	/**
	 * Adds the connector to the set of incoming connectors.
	 * 
	 * @param connector
	 *            connector to add as {@link BlockConnector}
	 */
	private BlockConnector addPreConnector(BlockConnector connector) {
		preActivityMap.put(connector.getTarget(), connector);
		return connector;
	}

	/**
	 * Adds a new outgoing {@link BlockConnector} for the {@link Activity} which
	 * is represented by this label.
	 * 
	 * @param label
	 *            label to add connection to
	 */
	public BlockConnector addPostConnector(String label) {
		return addPostConnector(Activity.getActivity(label));
	}

	/**
	 * Adds a new outgoing {@link BlockConnector} for the given activity to this
	 * block.
	 * 
	 * @param acivity
	 *            activity to connect to
	 */
	public BlockConnector addPostConnector(Activity acivity) {
		if (!postActivityMap.containsKey(acivity)) {
			BlockConnector connector = new BlockConnector(this, acivity);
			try {
				return addPostConnector(connector);
			} catch (UnexpectedException e) {
				LOG.debug("Connector " + connector + " was not added! "
						+ e.getMessage());
				e.printStackTrace();
				return postActivityMap.get(connector.getTarget());
			}
		} else {
			LOG.debug("Connector for activity " + acivity + " already exists.");
			return postActivityMap.get(acivity);
		}
	}

	/**
	 * Adds the connector to the set of outgoing connectors.
	 * 
	 * @param connector
	 *            connector to add as {@link BlockConnector}
	 * @throws UnexpectedException
	 */
	private BlockConnector addPostConnector(BlockConnector connector)
			throws UnexpectedException {
		if (!postActivityMap.containsKey(connector.getTarget())) {
			postActivityMap.put(connector.getTarget(), connector);
			return connector;
		} else {
			throw new UnexpectedException("Connector " + connector
					+ " already exits");
		}
	}

	/**
	 * This method removes the given outgoing connector from this block.
	 * 
	 * @param connector
	 */
	public void removePostConnector(BlockConnector connector) {
		postActivityMap.remove(connector.getTarget());
	}

	/**
	 * This method removes the given incoming connector of this block.
	 * 
	 * @param connector
	 *            connector as {@link BlockConnector}
	 */
	public void removePreConnector(BlockConnector connector) {
		preActivityMap.remove(connector.getTarget());
	}

	/**
	 * This method returns the outgoing connectors of this building block.
	 * 
	 * @return outgoing connectors as {@link Vector} of {@link BlockConnector}
	 */
	public Vector<BlockConnector> getPostConnectors() {
		Vector<BlockConnector> postConnectors = new Vector<BlockConnector>();

		postConnectors.addAll(postActivityMap.values());

		return postConnectors;
	}

	/**
	 * This method returns the incoming connectors of this building block.
	 * 
	 * @return incoming connectors as {@link Vector} of {@link BlockConnector}
	 */
	public Vector<BlockConnector> getPreConnectors() {
		Vector<BlockConnector> preConnectors = new Vector<BlockConnector>();

		preConnectors.addAll(preActivityMap.values());

		return preConnectors;
	}

	/**
	 * This method returns the {@link Activity} which is represented by this
	 * building block.
	 * 
	 * @return represented {@link Activity}
	 */
	public Activity getActivity() {
		return activity;
	}

	public void addToEpc(EPC epc) {
		HashMap<EPCNode, EPCNode> elementMapping = new HashMap<EPCNode, EPCNode>();

		// add function
		EPCFunction epcFunc = epc.addFunction(function.getLabel());
		elementMapping.put(function, epcFunc);

		for (EPCAndConnector and : getAndConnectors()) {
			EPCNode epcAnd = epc.addAndConnector();
			elementMapping.put(and, epcAnd);
		}

		for (BlockConnector connector : getPreConnectors()) {
			LooseEnd looseEnd = epc.addLooseEnd(activity,
					connector.getTarget(), false);
			elementMapping.put(connector, looseEnd);
		}

		for (BlockConnector connector : getPostConnectors()) {
			LooseEnd looseEnd = epc.addLooseEnd(activity,
					connector.getTarget(), true);
			elementMapping.put(connector, looseEnd);
		}

		for (EPCXorConnector xor : getXorConnectors()) {
			EPCXorConnector epcXor = epc.addXorConnector();
			elementMapping.put(xor, epcXor);
		}

		for (EPCEdge edge : getAllEdges()) {
			EPCNode start = elementMapping.get(edge.getStart());
			if (start == null) {
				LOG.error("Start " + edge.getStart() + " of block " + activity
						+ " has no mapping!");
				LOG.error("Start: Preset size: "
						+ edge.getStart().getPreset().size()
						+ ", postset size: "
						+ edge.getStart().getPostset().size());
				for (EPCEdge preEdge : edge.getStart().getPreset()) {
					LOG.error("Start: Preset: "
							+ preEdge.getStart()
							+ ", "
							+ preEdge.getStart().getPreset().firstElement()
									.getStart());
				}
				for (EPCEdge postEdge : edge.getStart().getPostset()) {
					LOG.error("Start: Postset: "
							+ postEdge.getEnd()
							+ ", "
							+ postEdge.getStart().getPostset().firstElement()
									.getEnd());
				}

				testEdges();
			}
			EPCNode end = elementMapping.get(edge.getEnd());
			if (end == null) {
				LOG.error("End " + edge.getEnd() + " has no mapping!");
			}
			epc.connect(start, end);
		}
	}

	private void testEdges() {
		Vector<EPCEdge> invalid = new Vector<EPCEdge>();
		Vector<EPCNode> missing = new Vector<EPCNode>();
		Vector<EPCNode> nodes = new Vector<EPCNode>();

		nodes.addAll(getFunctions());
		nodes.addAll(getAndConnectors());
		nodes.addAll(getXorConnectors());
		nodes.addAll(getPreConnectors());
		nodes.addAll(getPostConnectors());

		for (EPCEdge edge : getAllEdges()) {
			EPCNode start = edge.getStart();
			EPCNode end = edge.getEnd();

			if (!nodes.contains(start)) {
				missing.add(start);
				invalid.add(edge);
			}

			if (!nodes.contains(end)) {
				missing.add(start);
				invalid.add(edge);
			}
		}

		System.err.println("Invalid edges: " + invalid.size());
		System.err.println("Missing nodes: " + missing.size());

		for (EPCNode node : missing) {
			System.err.println("--\nMissing node: " + node);
			System.err.println("Preset: " + node.getPreset().size());
			System.err.println("Postset: " + node.getPostset().size());

			int max = (node.getPreset().size() > node.getPostset().size()) ? node
					.getPreset().size() : node.getPostset().size();

			for (int i = 0; i < max; i++) {
				String out = "";
				if (i < node.getPreset().size()) {
					out += node.getPreset().get(i).getStart();
				}

				out += " -- node -- ";

				if (i < node.getPostset().size()) {
					out += node.getPostset().get(i).getEnd();
				}
				System.err.println(out);
			}
		}

		for (EPCEdge edge : invalid) {
			System.err.println("invalid edge from " + edge.getStart() + " to "
					+ edge.getEnd() + ", edge: " + edge);
		}
	}

	@Override
	public boolean isNodeOfThisEpc(EPCNode node) {
		if (postActivityMap.containsValue(node)) {
			return true;
		} else if (preActivityMap.containsValue(node)) {
			return true;
		} else {
			return super.isNodeOfThisEpc(node);
		}
	}

	public boolean containsBlock(BuildingBlock block, boolean preset) {
		if (preset) {
			return containsBlockPreset(block);
		} else {
			return containsBlockPostset(block);
		}
	}

	private boolean containsBlockPreset(BuildingBlock otherBlock) {
		for (ActivitySet otherPreset : otherBlock.presets) {

			boolean found = false;

			for (ActivitySet thisPreset : presets) {

				LOG.info("Test if preset " + thisPreset + " contains "
						+ otherPreset + "? ");

				if (otherPreset.containsActivitySet(thisPreset)) {
					found = true;
					break;
				}
			}

			if (!found) {
				return false;
			}
		}

		return true;
	}

	private boolean containsBlockPostset(BuildingBlock otherBlock) {
		for (ActivitySet otherPostset : otherBlock.postsets) {

			boolean found = false;

			for (ActivitySet thisPostset : postsets) {

				LOG.info("Test if postset " + thisPostset + "contains "
						+ otherPostset + "? ");

				if (thisPostset.containsActivitySet(otherPostset)) {
					found = true;
					break;
				}
			}

			if (!found) {
				return false;
			}
		}

		return true;
	}

	public int getPresetCount() {
		return presets.size();
	}

	public int getPostsetCount() {
		return postsets.size();
	}
}
