package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.exec;

import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCFunction;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCNode;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCXorConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Vertex;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.exec.LpoState;

public class EpcState {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(EpcState.class);

	private EPC epc;

	private Vector<EPCFunction> activeFunctions;

	private Vector<EPCConnector> activeConnectors;

	private Vector<EPCXorConnector> activeAlternatives;

	private Vector<Vertex> fireVertices;
	private Vector<Activity> fireSequence;

	private Vector<LpoState> enabledLpos;

	private Lpo run;

	private int level = 0;
	private int alternative = 0;
	private int alternativeCount = 0;

	private EPCFunction lastFiredFunction = null;

	public EpcState(EPC epc, int alternative, int alternativeCount,
			Vector<LpoState> enabledLpos) {
		this.epc = epc;

		this.enabledLpos = new Vector<LpoState>();
		this.enabledLpos.addAll(enabledLpos);

		this.alternative = alternative;
		this.alternativeCount = alternativeCount;

		epc.markWithInitialMarking();

		Vector<Activity> startActivities = new Vector<Activity>();

		lastFiredFunction = null;

		for (EPCEdge edge : epc.getAllEdges()) {
			if (edge.isMarked()) {
				EPCNode node = edge.getStart();
				if (node instanceof EPCFunction) {
					Activity nodeActivity = Activity.getActivity(node
							.getLabel());
					if (!startActivities.contains(nodeActivity)) {
						startActivities.add(nodeActivity);
					}

					lastFiredFunction = (EPCFunction) node;
				} else {
					LOG.error("Unexpected start node " + node);
				}
			}
		}

		activeFunctions = new Vector<EPCFunction>();
		activeConnectors = new Vector<EPCConnector>();
		activeAlternatives = new Vector<EPCXorConnector>();

		fireSequence = new Vector<Activity>();
		fireVertices = new Vector<Vertex>();

		run = new Lpo("Run of " + toString());
		if (startActivities.size() != 1) {
			LOG.error("Unexpected start activities: " + startActivities);
		} else {
			fireSequence.add(startActivities.firstElement());
			fireVertices.add(run.addVertex(startActivities.firstElement()));
		}

		calcActive();
	}

	private EpcState() {
		lastFiredFunction = null;

		enabledLpos = new Vector<LpoState>();

		activeFunctions = new Vector<EPCFunction>();
		activeConnectors = new Vector<EPCConnector>();
		activeAlternatives = new Vector<EPCXorConnector>();

		fireSequence = new Vector<Activity>();
		fireVertices = new Vector<Vertex>();
	}

	public Lpo getRun() {
		return run;
	}

	public Vector<Activity> getFireSequence() {
		return fireSequence;
	}

	public Vertex getVertexForStep(int i) {
		return fireVertices.get(i);
	}

	public EPCFunction getLastFiredFunction() {
		return lastFiredFunction;
	}

	private void calcActive() {
		activeConnectors.clear();
		activeFunctions.clear();
		activeAlternatives.clear();

		Vector<EPCNode> active = new Vector<EPCNode>();

		for (EPCEdge edge : epc.getAllEdges()) {
			if (edge.isMarked()) {
				LOG.debug(this + ": Edge " + edge + " is marked.");
				if (edge.getEnd().canFire()) {
					active.add(edge.getEnd());
				}
			}
		}

		for (EPCNode activeNode : active) {
			if (activeNode instanceof EPCXorConnector) {
				if (activeNode.getPostset().size() > 1) {
					activeAlternatives.add((EPCXorConnector) activeNode);
					continue;
				} else {
					activeConnectors.add((EPCConnector) activeNode);
				}
			} else if (activeNode instanceof EPCConnector) {
				activeConnectors.add((EPCConnector) activeNode);
			} else {
				activeFunctions.add((EPCFunction) activeNode);
			}
		}
	}

	public EPC getEpc() {
		return epc;
	}

	public void fire(EPCNode node, EPCNode out) throws Exception {

		LOG.debug(this + ": firing " + node);

		Vector<Vertex> markSources = new Vector<Vertex>();
		if (node instanceof EPCFunction) {
			if (node.getPreset().size() > 1) {
				throw new Exception(
						"A function can only have one node in preset!");
			}

			EPCEdge edge = node.getPreset().firstElement();

			if (edge.getMarkSources() == null) {
				throw new Exception(
						"A function can only fire if the incoming edge is marked!");
			}

			// find pre vertices
			for (EPCFunction function : edge.getMarkSources()) {
				Activity functionActivity = Activity.getActivity(function
						.getLabel());
				int vertexIndex = fireSequence.lastIndexOf(functionActivity);
				markSources.add(fireVertices.get(vertexIndex));

			}

			// get activity for fired function
			Activity activity = Activity.getActivity(node.getLabel());
			// create vertex in run lpo for fired activity
			Vertex vertex = run.addVertex(activity);

			lastFiredFunction = (EPCFunction) node;

			// update fire sequences
			fireSequence.add(activity);
			fireVertices.add(vertex);

			// update run lpo
			for (Vertex markSource : markSources) {
				run.connectVertices(markSource, vertex);
			}
		}

		node.fire(out);

		calcActive();
	}

	public EpcState clone(int alternative, int alternativeCount) {
		EpcState clone = new EpcState();

		clone.run = run.clone();

		clone.epc = this.epc.clone();

		clone.alternative = alternative;

		clone.alternativeCount = alternativeCount;

		for (LpoState enabledLpo : enabledLpos) {
			clone.enabledLpos.add(enabledLpo.clone());
		}

		clone.fireSequence.clear();
		clone.fireSequence.addAll(fireSequence);

		clone.fireVertices.clear();

		for (Vertex vertex : fireVertices) {
			clone.fireVertices.add(run.translateVertexToClonedVertex(vertex));
		}

		clone.lastFiredFunction = (EPCFunction) clone.getEpc()
				.translateToClonedNode(lastFiredFunction);

		clone.level = this.level + 1;

		clone.calcActive();

		return clone;
	}

	public Vector<EPCFunction> getActiveFunctions() {
		return activeFunctions;
	}

	public Vector<EPCConnector> getActiveConnectors() {
		return activeConnectors;
	}

	public Vector<EPCXorConnector> getActiveAlternatives() {
		return activeAlternatives;
	}

	@Override
	public String toString() {
		return "EPC-" + level + " (alternatives: " + activeAlternatives.size()
				+ ", conn: " + activeConnectors.size() + ", " + activeFunctions
				+ ")";
	}

	public String shortName() {
		return "EPC-" + level + " (" + alternative + "/" + alternativeCount
				+ ")";
	}

	public Vector<LpoState> getEnabledLpos() {
		return enabledLpos;
	}

	public void setEnabledLpos(Vector<LpoState> enabledLpos) {
		this.enabledLpos = enabledLpos;
	}
}
