package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo;

import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class represents a labeled partial order (lpo). By convention each lpo
 * should start with a start vertex and end with an end vertex.
 * 
 * @author thomas
 * 
 */
public class Lpo {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(Lpo.class);

	/**
	 * label of the start vertex
	 */
	public static final String START_LABEL = "ST";

	/**
	 * label of the end vertex
	 */
	public static final String FINAL_LABEL = "FI";

	/**
	 * activities of this lpo
	 */
	private HashMap<String, Activity> activities;

	/**
	 * nodes of this lpo
	 */
	private Vector<Vertex> vertices;

	/**
	 * start vertex of this lpo
	 */
	private Vertex startVertex;

	/**
	 * end vertex of this lpo
	 */
	private Vertex finalVertex;

	/**
	 * name of this lpo for idenfification like file name, ...
	 */
	private String name;

	private HashMap<Vertex, Vertex> mapping;

	/**
	 * Constructs a new lpo and adds start and final vertex.
	 * 
	 * @param name
	 *            name of the lpo as {@link String}
	 */
	public Lpo(String name) {
		this.name = name;

		vertices = new Vector<Vertex>();
		activities = new HashMap<String, Activity>();

		Activity start = Activity.getActivity(Lpo.START_LABEL);
		activities.put(START_LABEL, start);
		startVertex = start.getVertex();
		startVertex.setLpo(this);
		vertices.add(startVertex);

		Activity fin = Activity.getActivity(Lpo.FINAL_LABEL);
		activities.put(FINAL_LABEL, fin);
		finalVertex = fin.getVertex();
		finalVertex.setLpo(this);
		vertices.add(finalVertex);

	}

	public Lpo(Vertex startVertex) {
		this.name = "aLpo";

		vertices = new Vector<Vertex>();
		activities = new HashMap<String, Activity>();

		Activity start = Activity.getActivity(Lpo.START_LABEL);
		activities.put(START_LABEL, start);
		this.startVertex = startVertex;
		vertices.add(startVertex);

		Activity fin = Activity.getActivity(Lpo.FINAL_LABEL);
		activities.put(FINAL_LABEL, fin);
		finalVertex = fin.getVertex();
		vertices.add(finalVertex);
	}

	/**
	 * This method connects all vertices with empty preset with the start vertex
	 * and all vertices with empty postset with the final vertex.
	 */
	public void finishConstruction() {
		Vector<Vertex> verticesWithEmptyPreset = new Vector<Vertex>();

		for (Vertex vertex : vertices) {
			if (vertex.getPreset().isEmpty()) {
				verticesWithEmptyPreset.add(vertex);
			}
		}

		// ignore start vertex
		verticesWithEmptyPreset.remove(startVertex);

		for (Vertex vertex : verticesWithEmptyPreset) {
			connectVertices(startVertex, vertex);
		}

		Vector<Vertex> verticesWithEmptyPostset = new Vector<Vertex>();

		for (Vertex vertex : vertices) {
			if (vertex.getPostset().isEmpty()) {
				verticesWithEmptyPostset.add(vertex);
			}
		}

		// ignore final vertex
		verticesWithEmptyPostset.remove(finalVertex);

		for (Vertex vertex : verticesWithEmptyPostset) {
			connectVertices(vertex, finalVertex);
		}

		// calculate skeleton edges
		calcSkeletonRecursive(startVertex);

		// add closure edges
		calcTransitiveClosureRecursive(startVertex);
	}

	private void calcTransitiveClosureRecursive(Vertex start) {

		calcTransitiveClosureEdges(start);

		for (LpoEdge edge : start.getPostset()) {
			calcTransitiveClosureRecursive(edge.getEnd());
		}
	}

	private void calcTransitiveClosureEdges(Vertex vertex) {
		Vector<Vertex> postPostVertices = new Vector<Vertex>();

		for (LpoEdge edge : vertex.getPostset()) {
			postPostVertices.addAll(getPostVertices(edge.getEnd()));
		}

		for (Vertex v : postPostVertices) {
			LpoEdge transitiveEdge = vertex.addTransitivPostVertex(v);
			v.addTransitivPreEdge(transitiveEdge);
		}
	}

	private void calcSkeletonRecursive(Vertex start) {
		LOG.debug("calc skeleton recursive of " + start.getActivity());

		calcPostSkeletonEdges(start);

		for (LpoEdge edge : start.getPostset()) {
			LOG.debug("calc skeleton recursive of " + start.getActivity()
					+ ": post node " + edge.getEnd().getActivity().getLabel());
			calcSkeletonRecursive(edge.getEnd());
		}
	}

	private void calcPostSkeletonEdges(Vertex vertex) {
		Vector<Vertex> postPostVertices = new Vector<Vertex>();

		LOG.debug("calc post skeleton edges of "
				+ vertex.getActivity().getLabel());

		for (LpoEdge edge : vertex.getPostset()) {
			postPostVertices.addAll(getPostVertices(edge.getEnd()));
		}

		for (LpoEdge edge : vertex.getPostset()) {
			if (postPostVertices.contains(edge.getEnd())) {
				edge.setSkeletonEdge(false);
				edge.setFoldingEdge(false);
			} else {
				edge.setSkeletonEdge(true);
				edge.setFoldingEdge(true);
			}
		}
	}

	private Vector<Vertex> getPostVertices(Vertex vertex) {

		LOG.debug("get post vertices of " + vertex.getActivity().getLabel());
		Vector<Vertex> postVertices = new Vector<Vertex>();

		for (LpoEdge edge : vertex.getPostset()) {

			postVertices.add(edge.getEnd());

			postVertices.addAll(getPostVertices(edge.getEnd()));
		}

		return postVertices;
	}

	/**
	 * This method returns all activities which are contained in this lpo.
	 * 
	 * @return {@link Activity}s of thie {@link Lpo}
	 */
	public Vector<Activity> getActivities() {
		Vector<Activity> activities = new Vector<Activity>();

		activities.addAll(this.activities.values());

		return activities;
	}

	/**
	 * This method search all vertices with the given label in this lpo.
	 * 
	 * @param label
	 *            label as {@link String}
	 * @return vertices which belongs to this {@link Lpo} and are labeled with
	 *         the given {@link String}
	 */
	public Vector<Vertex> getVertices(String label) {
		if (label == null)
			return null;

		Vector<Vertex> labelVertices = new Vector<Vertex>();

		for (Vertex vertex : this.vertices) {
			LOG.debug("Testing if " + vertex.getActivity().getLabel()
					+ " equals " + label);
			if (label.equals(vertex.getActivity().getLabel())) {
				labelVertices.add(vertex);
			}
		}

		return labelVertices;
	}

	/**
	 * This method returns the set of all vertices of this vertex.
	 * 
	 * @return {@link Vector} of {@link Vertex}
	 */
	public Vector<Vertex> getAllVertices() {
		return vertices;
	}

	/**
	 * This method disconnects the given vertices.
	 * 
	 * @param start
	 *            {@link Vertex} where the edge starts
	 * @param end
	 *            {@link Vertex} where the edge ends
	 */
	public void disconnectVertices(Vertex start, Vertex end) {
		start.removeVertexFromPostset(end);
		end.removeVertexFromPreset(start);
	}

	/**
	 * This method connects the given vertices.
	 * 
	 * @param start
	 *            {@link Vertex} where the edge starts
	 * @param end
	 *            {@link Vertex} where the edge ends
	 */
	public LpoEdge connectVertices(Vertex start, Vertex end) {
		if (start.getLpo() != this) {
			LOG.error("Start is no vertex of this lpo!");
			try {
				throw new Exception("Start is no vertex of this lpo!");
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		}

		if (end.getLpo() != this) {
			LOG.error("End is no vertex of this lpo!");
			try {
				throw new Exception("End is no vertex of this lpo!");
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		}

		if (!vertices.contains(start)) {
			LOG.error("Start is not known for this lpo!");
			try {
				throw new Exception("Start is not known for this lpo!");
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		}

		if (!vertices.contains(end)) {
			LOG.error("End is not known for this lpo!");
			try {
				throw new Exception("End is not known for this lpo!");
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		}

		LpoEdge edge = start.addVertexToPostset(end);
		end.addEdgeToPreset(edge);
		return edge;
	}

	/**
	 * This method removes the given vertex form this lpo.
	 * 
	 * @param vertex
	 *            {@link Vertex} to remove.
	 */
	public void removeVertex(Vertex vertex) {
		Vector<Vertex> preVertices = new Vector<Vertex>();

		for (LpoEdge pre : vertex.getPreset()) {
			preVertices.add(pre.getStart());
		}

		for (Vertex pre : preVertices) {
			disconnectVertices(pre, vertex);
		}

		Vector<Vertex> postVertices = new Vector<Vertex>();

		for (LpoEdge post : vertex.getPostset()) {
			postVertices.add(post.getEnd());
		}

		for (Vertex post : postVertices) {
			disconnectVertices(vertex, post);
		}

		vertices.remove(vertex);

		vertex.setLpo(null);
	}

	/**
	 * This method creates a new vertex with the given label and adds it to this
	 * lpo.
	 * 
	 * @param label
	 *            label of the vertex as {@link String}
	 * 
	 * @return created {@link Vertex}
	 */
	public Vertex addVertex(String label) {
		Activity activity = Activity.getActivity(label);

		if (!activities.containsKey(label)) {
			activities.put(label, activity);
		}

		return addVertex(activity);
	}

	/**
	 * This method creates a new vertex which represents the given activity and
	 * adds it to this lpo. If the activity is the LPO start label the start
	 * vertex is returned.
	 * 
	 * @param activity
	 *            {@link Activity} which is represented by this {@link Vertex}
	 * 
	 * @return created {@link Vertex}
	 */
	public Vertex addVertex(Activity activity) {
		if (activity.getLabel().equals(Lpo.START_LABEL)) {
			return startVertex;
		}

		if (!activities.containsKey(activity.getLabel())) {
			activities.put(activity.getLabel(), activity);
		}

		Vertex vertex = activity.getVertex();

		addVertex(vertex);

		return vertex;
	}

	public void addVertex(Vertex vertex) {
		if (!activities.containsKey(vertex.getActivity().getLabel())) {
			activities.put(vertex.getActivity().getLabel(),
					vertex.getActivity());
		}

		vertex.setLpo(this);

		vertices.add(vertex);
	}

	/**
	 * This method returns the start vertex of this lpo.
	 * 
	 * @return start vertex as {@link Vertex}
	 */
	public Vertex getStartVertex() {
		return startVertex;
	}

	/**
	 * This method returns the final vertex of this lpo.
	 * 
	 * @return final vertex as {@link Vertex}
	 */
	public Vertex getFinalVertex() {
		return finalVertex;
	}

	/**
	 * This method returns the name of this lpo.
	 * 
	 * @return name as {@link String}
	 */
	public String getName() {
		return name;
	}

	@Override
	public String toString() {
		return "lpo " + name + " (" + hashCode() + ")";
	}

	public Lpo clone() {
		Lpo clone = new Lpo(name);

		clone.activities.clear();
		clone.activities.putAll(activities);

		mapping = new HashMap<Vertex, Vertex>();

		for (Vertex vertex : vertices) {
			if (vertex == startVertex) {
				mapping.put(startVertex, clone.startVertex);
				continue;
			}

			if (vertex == finalVertex) {
				mapping.put(finalVertex, clone.finalVertex);
				continue;
			}

			Vertex clonedVertex = clone.addVertex(vertex.getActivity());

			mapping.put(vertex, clonedVertex);
		}

		for (Vertex vertex : vertices) {
			Vertex cloneVertex = mapping.get(vertex);
			for (LpoEdge edge : vertex.getPostset()) {
				Vertex cloneEnd = mapping.get(edge.getEnd());
				clone.connectVertices(cloneVertex, cloneEnd);
			}
		}

		return clone;
	}

	public Vertex translateVertexToClonedVertex(Vertex vertex) {
		return mapping.get(vertex);
	}

	public int distance(Activity start, Activity end) {

		Vector<Vertex> vertices = getVertices(start.getLabel());

		int min = Integer.MAX_VALUE;

		for (Vertex vertex : vertices) {
			int dist = distance(vertex, end, 1);
			if (dist < min) {
				min = dist;
			}
		}

		return min;
	}

	private int distance(Vertex start, Activity end, int distance) {

		Vector<LpoEdge> skelettonPost = start.getSkeletonPostset();

		Vector<Vertex> vertices = new Vector<Vertex>();

		for (LpoEdge edge : skelettonPost) {

			Vertex ee = edge.getEnd();
			if (ee.getActivity().equals(end)) {
				return distance;
			} else {
				vertices.add(ee);
			}

		}

		int min = Integer.MAX_VALUE;

		for (Vertex v : vertices) {

			int dist = distance(v, end, distance + 1);
			if (dist < min) {
				min = dist;
			}

		}

		return min;
	}

}
