package de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo;

import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A vertex is a node in an LPO graph.
 * 
 * @author thomas
 * 
 */
public class Vertex {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(Activity.class);

	/**
	 * {@link Activity} which is represented by this vertex
	 */
	private Activity activity;

	/**
	 * preset of this vertex (all incoming edges of this vertex)
	 */
	private Vector<LpoEdge> preset;

	/**
	 * postset of this vertex (all outgoing edges of this vertex)
	 */
	private Vector<LpoEdge> postset;

	/**
	 * reference to the lpo where this vertex belongs to.
	 */
	private Lpo lpo;

	/**
	 * Constructs a new {@link Vertex} for the given {@link Activity}.
	 * 
	 * @param activity
	 *            {@link Activity} represented by this {@link Vertex}
	 */
	public Vertex(Activity activity) {
		this.activity = activity;

		// add vertex to activity
		activity.addVertex(this);

		LOG.debug("New instance of " + activity.getLabel());

		preset = new Vector<LpoEdge>();
		postset = new Vector<LpoEdge>();
	}

	/**
	 * This method sets the lpo where this vertex belongs to.
	 * 
	 * @param lpo
	 *            lpo as {@link Lpo}
	 */
	public void setLpo(Lpo lpo) {
		if (this.lpo != null && lpo != null) {
			LOG.error("The vertex " + toString()
					+ " was already added to the lpo " + this.lpo.toString()
					+ " and is now added to the lpo " + lpo.toString());
		}

		this.lpo = lpo;
	}

	/**
	 * This method returns the lpo where this vertex belongs to.
	 * 
	 * @return lpo as {@link Lpo}
	 */
	public Lpo getLpo() {
		return lpo;
	}

	/**
	 * This method returns the {@link Activity} this {@link Vertex} represents.
	 * 
	 * @return represented {@link Activity}
	 */
	public Activity getActivity() {
		return activity;
	}

	/**
	 * This method returns the preset of this vertex. (Incoming edges of this
	 * vertex.)
	 * 
	 * @return list of incoming edges of this vertex
	 */
	public Vector<LpoEdge> getPreset() {
		return preset;
	}

	/**
	 * This method returns the postset of this vertex. (Outgoing edges of this
	 * vertex.)
	 * 
	 * @return list of outgoing edges of this vertex
	 */
	public Vector<LpoEdge> getPostset() {
		return postset;
	}

	/**
	 * This method returns all incoming skeleton edges of this vertex.
	 * 
	 * @return skeleton preset
	 */
	public Vector<LpoEdge> getSkeletonPreset() {
		Vector<LpoEdge> skeletonPreset = new Vector<LpoEdge>();

		for (LpoEdge pre : preset) {
			if (pre.isSkeletonEdge()) {
				skeletonPreset.add(pre);
			}
		}

		return skeletonPreset;
	}

	/**
	 * This method returns the activities of the vertices which are connected
	 * with incoming skeleton edges with this vertex.
	 * 
	 * @return predecessor set of {@link Activity}
	 */
	public Vector<Activity> getPredecessorSet() {
		Vector<Activity> predecessorSet = new Vector<Activity>();

		for (LpoEdge preSkeleton : getSkeletonPreset()) {
			predecessorSet.add(preSkeleton.getStart().getActivity());
		}

		return predecessorSet;
	}

	/**
	 * This method returns all activities which happen before this vertex.
	 * 
	 * @return activities before this vertex as {@link Activity}
	 */
	public Vector<Activity> getPreActivities() {
		Vector<Activity> preActivities = new Vector<Activity>();

		Vector<Vertex> preVertices = getPreVertices();
		for (Vertex pre : preVertices) {
			preActivities.add(pre.getActivity());
		}

		return preActivities;
	}

	/**
	 * This method returns all vertices which are transitive ordered before this
	 * vertex.
	 * 
	 * @return vertices before this vertex
	 */
	public Vector<Vertex> getPreVertices() {
		Vector<Vertex> vertices = new Vector<Vertex>();

		for (LpoEdge edge : preset) {
			if (!vertices.contains(edge)) {
				vertices.add(edge.getStart());
			}
		}

		return vertices;
	}

	/**
	 * This method returns all vertices which are transitive ordered after this
	 * vertex.
	 * 
	 * @return vertices after this vertex
	 */
	public Vector<Vertex> getPostVertices() {
		Vector<Vertex> vertices = new Vector<Vertex>();

		for (LpoEdge vertex : postset) {
			vertices.add(vertex.getEnd());
		}

		return vertices;
	}

	/**
	 * This method returns all outgoing skeleton edges of this vertex.
	 * 
	 * @return skeleton postset
	 */
	public Vector<LpoEdge> getSkeletonPostset() {
		Vector<LpoEdge> skeletonPostset = new Vector<LpoEdge>();

		for (LpoEdge post : postset) {
			if (post.isSkeletonEdge()) {
				skeletonPostset.add(post);
			}
		}

		return skeletonPostset;
	}

	/**
	 * This method returns the activities of the vertices which are connected
	 * with outgoing skeleton edges with this vertex.
	 * 
	 * @return successor set of {@link Activity}
	 */
	public Vector<Activity> getSuccessorSet() {
		Vector<Activity> successorSet = new Vector<Activity>();

		for (LpoEdge postSkeleton : getSkeletonPostset()) {
			successorSet.add(postSkeleton.getEnd().getActivity());
		}

		return successorSet;
	}

	/**
	 * This method returns all activities which happen after this vertex.
	 * 
	 * @return activities after this vertex as {@link Activity}
	 */
	public Vector<Activity> getPostActivities() {
		Vector<Activity> postActivities = new Vector<Activity>();

		Vector<Vertex> postVertices = getPostVertices();
		for (Vertex post : postVertices) {
			postActivities.add(post.getActivity());
		}

		return postActivities;
	}

	/**
	 * This method adds the given vertex to the preset of this vertex.
	 * 
	 * @param vertex
	 *            vertex to add to preset
	 */
	public LpoEdge addVertexToPreset(Vertex vertex) {
		for (LpoEdge edge : preset) {
			if (edge.getEnd().equals(vertex)) {
				LOG.debug("Vertex " + vertex.getActivity().getLabel()
						+ " is already part of the preset.");
				return edge;
			}
		}

		LpoEdge edge = new LpoEdge(vertex, this, true);
		preset.add(edge);

		vertex.addEdgeToPostset(edge);

		return edge;
	}

	/**
	 * This method removes the given vertex form the preset of this vertex.
	 * 
	 * @param vertex
	 *            vertex to remove from preset
	 */
	public LpoEdge removeVertexFromPreset(Vertex vertex) {
		LpoEdge edge = null;

		for (LpoEdge pre : preset) {
			if (pre.getStart().equals(vertex)) {
				edge = pre;
				break;
			}
		}

		if (edge != null) {
			preset.remove(edge);
			vertex.removeVertexFromPostset(this);
		}

		return edge;
	}

	/**
	 * This method adds the given vertex to the postset of this vertex.
	 * 
	 * @param vertex
	 *            vertex to add to postset
	 */
	public LpoEdge addVertexToPostset(Vertex vertex) {
		for (LpoEdge edge : postset) {
			if (edge.getEnd().equals(vertex)) {
				LOG.debug("Vertex " + vertex.getActivity().getLabel()
						+ " is already part of the postset.");
				return edge;
			}
		}

		LpoEdge edge = new LpoEdge(this, vertex, true);
		postset.add(edge);

		vertex.addEdgeToPreset(edge);

		return edge;
	}

	public void addEdgeToPostset(LpoEdge edge) {
		postset.add(edge);
	}

	public void addEdgeToPreset(LpoEdge edge) {
		for (LpoEdge preEdge : preset) {
			if (preEdge.getStart().equals(edge.getStart())) {
				LOG.debug("Edge to " + edge.getStart().getActivity().getLabel()
						+ " already exists.");
				return;
			}
		}

		preset.add(edge);
	}

	public LpoEdge addTransitivPostVertex(Vertex vertex) {
		for (LpoEdge edge : postset) {
			if (edge.getEnd().equals(vertex)) {
				LOG.debug("Edge to " + vertex.getActivity().getLabel()
						+ " already exists.");
				return edge;
			}
		}

		LpoEdge edge = new LpoEdge(this, vertex, false);
		postset.add(edge);

		return edge;
	}

	public void addTransitivPreEdge(LpoEdge transitiveEdge) {
		for (LpoEdge edge : preset) {
			if (edge.getStart().equals(transitiveEdge.getStart())) {
				LOG.debug("Edge to "
						+ transitiveEdge.getStart().getActivity().getLabel()
						+ " already exists.");
				return;
			}
		}

		preset.add(transitiveEdge);
	}

	/**
	 * This method removes the given vertex form the postset of this vertex.
	 * 
	 * @param vertex
	 *            vertex to remove from postset
	 */
	public LpoEdge removeVertexFromPostset(Vertex vertex) {
		LpoEdge edge = null;

		for (LpoEdge post : postset) {
			if (post.getEnd().equals(vertex)) {
				edge = post;
				break;
			}
		}

		if (edge != null) {
			postset.remove(edge);
			vertex.removeVertexFromPreset(this);
		}

		return edge;
	}

	@Override
	public String toString() {
		return "Instance of " + activity.getLabel();
	}
}
