package de.fernunihagen.viptool.algorithm.folding.epcoflpo.io.epml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCAndConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCFunction;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCNode;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCXorConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;

/**
 * This class transforms the internal EPC structure to an epml file.
 * 
 * @author thomas
 * 
 */
public class EPMLDocumentBuilder {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory
			.getLogger(EPMLDocumentBuilder.class);

	/**
	 * IdAssigner for unique element ids
	 */
	private IdAssigner idAssigner = null;

	/**
	 * epml document
	 */
	private Document document = null;

	/**
	 * pointer to the document root
	 */
	private Element epmlElement = null;

	/**
	 * pointer to the current epc element
	 */
	private Element currentEPCElement = null;

	/**
	 * node type to epml name mapping
	 */
	private Map<Class<? extends EPCNode>, String> classMappings;

	/**
	 * Constructs a new document builder and assigns the names for the epc
	 * elements and prepares the document.
	 */
	public EPMLDocumentBuilder() {
		classMappings = new HashMap<Class<? extends EPCNode>, String>();

		classMappings.put(EPCFunction.class, "function");
		classMappings.put(EPCXorConnector.class, "xor");
		classMappings.put(EPCAndConnector.class, "and");

		idAssigner = new IdAssigner();

		epmlElement = null;
		currentEPCElement = null;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = null;
		try {
			docBuilder = factory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			LOG.error("Document couldn't be created! " + e.getMessage());
			e.printStackTrace();
			return;
		}
		document = docBuilder.newDocument();
		document.setXmlVersion("1.0");
		document.setXmlStandalone(true);

		epmlElement = buildEPMLElement();
		document.appendChild(epmlElement);
	}

	/**
	 * Returns the epml document.
	 * 
	 * @return epml document as {@link Document}
	 */
	public Document toDocument() {
		return document;
	}

	/**
	 * This method set the epml node up. This means the node is created und some
	 * required attributes like namespace are set.
	 * 
	 * @return epml root node as {@link Element}
	 * 
	 */
	private Element buildEPMLElement() {
		Element epmlElement = document.createElementNS("http://www.epml.de",
				"epml:epml");
		epmlElement.setAttribute("xmlns:xsi",
				"http://www.w3.org/2001/XMLSchema-instance");
		epmlElement.setAttribute("xsi:schemaLocation", "epml_1_draft.xsd");

		return epmlElement;
	}

	/**
	 * This method adds an EPC node to the document.
	 * 
	 * @param epc
	 *            epc to add as {@link EPC}
	 */
	public void addElement(EPC epc) {
		currentEPCElement = document.createElement("epc");
		currentEPCElement.setAttribute("EpcId",
				String.valueOf(idAssigner.lookup(epc)));

		epmlElement.appendChild(currentEPCElement);
	}

	/**
	 * Adds a data node to the document.
	 * 
	 * @param node
	 *            node as {@link EPCNode}
	 */
	public void addElement(EPCNode node) {
		if (node.getLabel().equals(Lpo.START_LABEL)) {
			addElementWithNameChildElement("start", idAssigner.lookup(node),
					node.getLabel());
		} else if (node.getLabel().equals(Lpo.FINAL_LABEL)) {
			addElementWithNameChildElement("final", idAssigner.lookup(node),
					node.getLabel());
		} else {
			addElementWithNameChildElement(classMappings.get(node.getClass()),
					idAssigner.lookup(node), node.getLabel());
		}
	}

	/**
	 * Adds an EPCEdge to the document.
	 * 
	 * @param edge
	 *            edge as {@link EPCEdge}
	 */
	public void addElement(EPCEdge edge) {
		Element arcElement = document.createElement("arc");

		arcElement.setAttribute("id", String.valueOf(idAssigner.lookup(edge)));

		Element flowElement = document.createElement("flow");

		flowElement.setAttribute("source",
				String.valueOf(idAssigner.lookup(edge.getStart())));

		flowElement.setAttribute("target",
				String.valueOf(idAssigner.lookup(edge.getEnd())));

		arcElement.appendChild(flowElement);

		currentEPCElement.appendChild(arcElement);
	}

	/**
	 * Creates a new node element and adds is as child of the current epc
	 * element.
	 * 
	 * @param elementName
	 *            node name of the element
	 * @param elementId
	 *            id of the element
	 * @param nameContent
	 *            name of the element (name attribute)
	 */
	private void addElementWithNameChildElement(String elementName,
			int elementId, String nameContent) {
		Element element = document.createElement(elementName);
		element.setAttribute("id", String.valueOf(elementId));

		Element nameElement = document.createElement("name");
		nameElement.setTextContent(nameContent);
		element.appendChild(nameElement);

		currentEPCElement.appendChild(element);
	}

	public static void saveEpcAsFile(EPC epc, File file) {
		EPMLDocumentBuilder docbuilder = new EPMLDocumentBuilder();

		docbuilder.addElement(epc);

		for (EPCNode node : epc.getFunctions()) {
			docbuilder.addElement(node);
		}

		for (EPCNode node : epc.getAndConnectors()) {
			docbuilder.addElement(node);
		}

		for (EPCNode node : epc.getXorConnectors()) {
			docbuilder.addElement(node);
		}

		for (EPCEdge edge : epc.getAllEdges()) {
			docbuilder.addElement(edge);
		}

		storeXMLFile(file, docbuilder.toDocument());
	}

	private static void storeXMLFile(File file, Document document) {
		Writer writer;
		Transformer transformer;
		final TransformerFactory factory = TransformerFactory.newInstance();

		try {
			factory.setAttribute("indent-number", new Integer(4));
			transformer = factory.newTransformer();
			// set use whitespaces
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			// set output-type xml
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			// Content-Type
			transformer.setOutputProperty(OutputKeys.MEDIA_TYPE, "text/xml");
			// set encoding
			transformer.setOutputProperty(OutputKeys.ENCODING, "utf-8");

			// create file writer
			writer = new OutputStreamWriter(new FileOutputStream(file), Charset
					.forName("utf-8").newEncoder());

			transformer.transform(new DOMSource(document), new StreamResult(
					writer));
		} catch (Exception exception) {
			LOG.error("File save error ("
					+ exception.getClass().getCanonicalName() + "): "
					+ exception.getMessage());
			exception.printStackTrace();
		}
	}
}
