package de.fernunihagen.viptool.algorithm.folding.epcoflpo.io.lpo;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.LpoEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Vertex;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.io.epml.IdAssigner;

public class PNMLDocumentBuilder {

	public static void saveLpoAsFile(Lpo lpo, File file) {
		PNMLDocumentBuilder docbuilder = new PNMLDocumentBuilder();

		docbuilder.addElement(lpo);

		for (Vertex vertex : lpo.getAllVertices()) {
			docbuilder.addElement(vertex);
		}

		for (Vertex vertex : lpo.getAllVertices()) {
			docbuilder.addElement(vertex);
		}

		storeXMLFile(file, docbuilder.toDocument());
	}

	private static void storeXMLFile(File file, Document document) {
		Writer writer;
		Transformer transformer;
		final TransformerFactory factory = TransformerFactory.newInstance();

		try {
			factory.setAttribute("indent-number", new Integer(4));
			transformer = factory.newTransformer();
			// set use whitespaces
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			// set output-type xml
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			// Content-Type
			transformer.setOutputProperty(OutputKeys.MEDIA_TYPE, "text/xml");
			// set encoding
			transformer.setOutputProperty(OutputKeys.ENCODING, "utf-8");

			// create file writer
			writer = new OutputStreamWriter(new FileOutputStream(file), Charset
					.forName("utf-8").newEncoder());

			transformer.transform(new DOMSource(document), new StreamResult(
					writer));
		} catch (Exception exception) {
			LOG.error("File save error ("
					+ exception.getClass().getCanonicalName() + "): "
					+ exception.getMessage());
			exception.printStackTrace();
		}
	}

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory
			.getLogger(PNMLDocumentBuilder.class);

	/**
	 * unique id generator
	 */
	private IdAssigner idAssigner;

	/**
	 * XML document which is constructed
	 */
	private Document document;

	/**
	 * Pointer to the current element
	 */
	private Element pnmlElement;

	/**
	 * Reference to the lpo which is written at the moment.
	 */
	private Element currentLPOElement;

	/**
	 * Constructs a new {@link PNMLDocumentBuilder} and initializes the
	 * datastructure to XML mappings
	 */
	private PNMLDocumentBuilder() {
		idAssigner = new IdAssigner();
		pnmlElement = null;
		currentLPOElement = null;

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docBuilder = null;
		try {
			docBuilder = factory.newDocumentBuilder();

			document = docBuilder.newDocument();
			document.setXmlVersion("1.0");
			document.setXmlStandalone(false);

			pnmlElement = document.createElement("pnml");

			document.appendChild(pnmlElement);
		} catch (ParserConfigurationException e) {
			LOG.error("Document builder factory couldn't be created! "
					+ e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Returns the document of this {@link PNMLDocumentBuilder}
	 * 
	 * @return document as {@link Document}
	 */
	public Document toDocument() {
		return document;
	}

	/**
	 * This method creates a new LPO into this PNML document.
	 * 
	 * @param lpo
	 *            lpo which should be added
	 */
	public void addElement(Lpo lpo) {

		currentLPOElement = document.createElement("lpo");

		currentLPOElement.setAttribute("id", lpo.getName() + ".lpo");
		currentLPOElement.setAttribute("type", "VIPSchema.xsd");

		Element nameElement = document.createElement("name");
		Element valueElement = document.createElement("value");

		valueElement.setTextContent(lpo.getName());
		nameElement.appendChild(valueElement);

		currentLPOElement.appendChild(nameElement);

		pnmlElement.appendChild(currentLPOElement);
	}

	/**
	 * Adds a new event to the lpo.
	 * 
	 * @param node
	 *            event to add as {@link LPOEvent}
	 */
	public void addElement(Vertex node) {
		Element nodeElement = document.createElement("event");

		nodeElement.setAttribute("id", String.valueOf(idAssigner.lookup(node)));
		nodeElement.setAttribute("provider", "lpo.provider.event");

		Element nameElement = document.createElement("name");
		Element valueElement = document.createElement("value");

		valueElement.setTextContent(node.getActivity().getLabel());

		nameElement.appendChild(valueElement);
		nodeElement.appendChild(nameElement);

		currentLPOElement.appendChild(nodeElement);
	}

	/**
	 * Adds all outgoing arcs of this event to the lpo.
	 * 
	 * @param node
	 *            event as {@link LPOEvent}
	 */
	public void addOutgoingArcs(Vertex node) {
		for (LpoEdge edge : node.getPostset()) {
			addArc(node, edge.getEnd());
		}
	}

	/**
	 * This method ads an arc form the event fromNode to the event toNode to the
	 * lpo.
	 * 
	 * @param fromNode
	 *            start event of the arc as {@link LPOEvent}
	 * @param toNode
	 *            end event of the arc as {@link LPOEvent}
	 */
	private void addArc(Vertex fromNode, Vertex toNode) {
		Element arcElement = document.createElement("lpoArc");
		arcElement.setAttribute("id",
				String.valueOf(idAssigner.lookup(arcElement)));
		arcElement.setAttribute("provider", "lpo.provider.edge");
		arcElement.setAttribute("source",
				String.valueOf(idAssigner.lookup(fromNode)));
		arcElement.setAttribute("target",
				String.valueOf(idAssigner.lookup(toNode)));

		currentLPOElement.appendChild(arcElement);
	}

}
