package de.fernunihagen.viptool.algorithm.folding.epcoflpo.logic;

import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCNode;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.LooseEnd;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.buildingblocks.ActivitySet;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.buildingblocks.BuildingBlock;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.LpoEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Vertex;

public class EpcFolding {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory.getLogger(EpcFolding.class);

	private Vector<Lpo> specification;

	private HashMap<Activity, HashMap<Activity, LooseEnd>> unsolvedOutgoingLooseEnds;
	private HashMap<Activity, HashMap<Activity, LooseEnd>> unsolvedIncomingLooseEnds;

	public EpcFolding() {
		specification = new Vector<Lpo>();

		unsolvedOutgoingLooseEnds = new HashMap<Activity, HashMap<Activity, LooseEnd>>();
		unsolvedIncomingLooseEnds = new HashMap<Activity, HashMap<Activity, LooseEnd>>();
	}

	public void addToSpecification(Vector<Lpo> lpos) {
		specification.addAll(lpos);
	}

	public EPC runAlgo() {
		HashMap<Activity, BuildingBlock> blocks = generateBuildingBlocks(getSpecification());
		EPC epc = mergeBlocks(blocks);
		return epc;
	}

	public HashMap<Activity, BuildingBlock> generateBuildingBlocks(
			Vector<Lpo> specification) {
		HashMap<Activity, BuildingBlock> blocks = new HashMap<Activity, BuildingBlock>();

		// build blocks from specification
		for (Lpo lpo : specification) {
			addPreAndPostsetsToBlocks(blocks, lpo);
		}

		// remove useless connectors of all blocks
		for (BuildingBlock block : blocks.values()) {
			block.removeUselessConnectors();
		}

		return blocks;
	}

	private void merge(EPC epc, LooseEnd aLooseEnd, LooseEnd anotherLooseEnd) {

		if (!aLooseEnd.isValid() || !anotherLooseEnd.isValid()) {
			LOG.error("Can not merge " + aLooseEnd + " and " + anotherLooseEnd
					+ " because one of them is invalid!");
			return;
		}

		LooseEnd incomingLooseEnd = null;
		LooseEnd outgoingLooseEnd = null;

		if (aLooseEnd.getPreset().isEmpty()) {
			incomingLooseEnd = aLooseEnd;
			outgoingLooseEnd = anotherLooseEnd;
		} else {
			incomingLooseEnd = anotherLooseEnd;
			outgoingLooseEnd = aLooseEnd;
		}

		EPCNode edgeStart = outgoingLooseEnd.getPreset().firstElement()
				.getStart();
		EPCNode edgeEnd = incomingLooseEnd.getPostset().firstElement().getEnd();

		EPCEdge.deleteAllEdges(outgoingLooseEnd);
		EPCEdge.deleteAllEdges(incomingLooseEnd);

		epc.connect(edgeStart, edgeEnd);
	}

	private void fillLooseEndMaps(Vector<LooseEnd> looseEnds) throws Exception {
		for (LooseEnd looseEnd : looseEnds) {
			if (looseEnd.isOutgoingLooseEnd()) {
				if (!unsolvedOutgoingLooseEnds.containsKey(looseEnd.getStart())) {
					unsolvedOutgoingLooseEnds.put(looseEnd.getStart(),
							new HashMap<Activity, LooseEnd>());
				}

				if (unsolvedOutgoingLooseEnds.get(looseEnd.getStart())
						.containsKey(looseEnd.getEnd())) {
					throw new Exception("The loose end " + looseEnd
							+ " is already in the map!");
				} else {
					unsolvedOutgoingLooseEnds.get(looseEnd.getStart()).put(
							looseEnd.getEnd(), looseEnd);
				}
			} else {
				if (!unsolvedIncomingLooseEnds.containsKey(looseEnd.getStart())) {
					unsolvedIncomingLooseEnds.put(looseEnd.getStart(),
							new HashMap<Activity, LooseEnd>());
				}

				if (unsolvedIncomingLooseEnds.get(looseEnd.getStart())
						.containsKey(looseEnd.getEnd())) {
					throw new Exception("The loose end " + looseEnd
							+ " is already in the map!");
				} else {
					unsolvedIncomingLooseEnds.get(looseEnd.getStart()).put(
							looseEnd.getEnd(), looseEnd);
				}
			}
		}
	}

	public EPC mergeBlocks(HashMap<Activity, BuildingBlock> blocks) {
		EPC foldedEpc = new EPC();

		// add all blocks to epc
		for (BuildingBlock block : blocks.values()) {
			block.addToEpc(foldedEpc);
		}

		unsolvedIncomingLooseEnds.clear();
		unsolvedOutgoingLooseEnds.clear();

		try {
			fillLooseEndMaps(foldedEpc.getLooseEnds());
		} catch (Exception e) {
			LOG.error("Error while filling maps! " + e.getMessage());
			e.printStackTrace();
			return null;
		}

		LOG.debug("Merging blocks: Loose ends count: "
				+ foldedEpc.getLooseEnds().size());

		LOG.debug("Incoming loose ends: " + unsolvedIncomingLooseEnds.size());
		LOG.debug("Outgoing loose ends: " + unsolvedOutgoingLooseEnds.size());

		for (Activity start : unsolvedOutgoingLooseEnds.keySet()) {
			for (Activity end : unsolvedOutgoingLooseEnds.get(start).keySet()) {
				LooseEnd outgoing = unsolvedOutgoingLooseEnds.get(start).get(
						end);
				LooseEnd incoming = unsolvedIncomingLooseEnds.get(end).get(
						start);

				LOG.debug("Merging " + outgoing + " with " + incoming);
				merge(foldedEpc, outgoing, incoming);

				unsolvedIncomingLooseEnds.get(end).remove(start);
				if (unsolvedIncomingLooseEnds.get(end).isEmpty()) {
					unsolvedIncomingLooseEnds.remove(end);
				}
			}
		}

		if (!unsolvedIncomingLooseEnds.isEmpty()) {

			LOG.error("Bulding blocks doesn't match!");

			for (Activity startActivity : unsolvedIncomingLooseEnds.keySet()) {
				for (Activity endActivity : unsolvedIncomingLooseEnds.get(
						startActivity).keySet()) {
					LOG.error("Unsolved loose end "
							+ unsolvedIncomingLooseEnds.get(startActivity).get(
									endActivity));
				}
			}

			return null;
		}

		foldedEpc.removeUselessConnectors();

		return foldedEpc;
	}

	private void addPreAndPostsetsToBlocks(
			HashMap<Activity, BuildingBlock> blocks, Lpo lpo) {
		Vertex vertex = lpo.getStartVertex();

		addVertexToBlocks(blocks, vertex);

		addPreAndPostsetsRecursive(blocks,
				lpoEdgesToVertices(vertex.getSkeletonPostset(), false));
	}

	private Vector<Vertex> lpoEdgesToVertices(Vector<LpoEdge> edges,
			boolean start) {
		Vector<Vertex> vertices = new Vector<Vertex>();

		for (LpoEdge edge : edges) {
			if (start) {
				vertices.add(edge.getStart());
			} else {
				vertices.add(edge.getEnd());
			}
		}

		return vertices;
	}

	private void addPreAndPostsetsRecursive(
			HashMap<Activity, BuildingBlock> blocks, Vector<Vertex> vertices) {
		for (Vertex vertex : vertices) {
			addVertexToBlocks(blocks, vertex);
			addPreAndPostsetsRecursive(blocks,
					lpoEdgesToVertices(vertex.getSkeletonPostset(), false));
		}
	}

	public void addVertexToBlocks(HashMap<Activity, BuildingBlock> blocks,
			Vertex vertex) {

		if (!blocks.containsKey(vertex.getActivity())) {
			blocks.put(vertex.getActivity(),
					new BuildingBlock(vertex.getActivity()));
		}

		BuildingBlock block = blocks.get(vertex.getActivity());

		ActivitySet preset = new ActivitySet();
		for (LpoEdge edge : vertex.getPreset()) {
			if (edge.isFoldingEdge()) {
				preset.add(edge.getStart().getActivity());
			}
		}

		block.addPreset(preset);

		ActivitySet postset = new ActivitySet();
		for (LpoEdge edge : vertex.getPostset()) {
			if (edge.isFoldingEdge()) {
				postset.add(edge.getEnd().getActivity());
			}
		}

		block.addPostset(postset);
	}

	public Vector<Lpo> getSpecification() {
		return specification;
	}
}
