package de.fernunihagen.viptool.algorithm.folding.epcoflpo.logic;

import java.util.Collections;
import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.buildingblocks.BuildingBlock;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.LpoEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Vertex;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.exec.WrongContinuation;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.util.ActivityPair;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.util.ActivityPairDistanceComparator;

public class FindPossibleEdge {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory
			.getLogger(FindPossibleEdge.class);

	private EpcFolding foldingAlgo;
	private WrongContinuation wrongContinuation;

	public FindPossibleEdge(EpcFolding foldingAlgo,
			WrongContinuation wrongContinuation) {
		this.foldingAlgo = foldingAlgo;
		this.wrongContinuation = wrongContinuation;
	}

	public ActivityPair runAlgo() {

		Vector<ActivityPair> labelPairs = findPossibleLabelPairs(
				foldingAlgo.getSpecification(), wrongContinuation);

		for (ActivityPair pair : labelPairs) {
			if (testActivityPair(pair, wrongContinuation.getRun())) {
				LOG.info("Edges for activity pair " + pair
						+ " enabled and fixed the wrong continuation "
						+ wrongContinuation);
				return pair;
			}
		}

		LOG.info("It was no solution found which fix the wrong continuation "
				+ wrongContinuation);

		return null;
	}

	private boolean testActivityPair(ActivityPair pair, Lpo run) {
		Vector<LpoEdge> changedEdges = new Vector<LpoEdge>();

		LOG.info("Test activity pair " + pair);

		for (LpoEdge edge : pair.getPairEdges()) {
			if (!edge.isFoldingEdge()) {
				edge.setFoldingEdge(true);
				changedEdges.add(edge);
			}
		}

		Vector<LpoEdge> changedRunEdges = new Vector<LpoEdge>();
		for (Vertex vertex : run.getVertices(pair.getStart().getLabel())) {
			for (LpoEdge edge : vertex.getPostset()) {
				if (edge.getEnd().getActivity().equals(pair.getEnd())) {
					if (!edge.isFoldingEdge()) {
						edge.setFoldingEdge(true);
						changedRunEdges.add(edge);
					}
				}
			}
		}

		LOG.debug(changedEdges.size()
				+ " edges in wrong continuation run changed.");

		if (!changedEdges.isEmpty()) {

			HashMap<Activity, BuildingBlock> wrongContinuationBlocks = generateBlocksForWrongContinuation(run);

			HashMap<Activity, BuildingBlock> specBlocks = foldingAlgo
					.generateBuildingBlocks(foldingAlgo.getSpecification());

			BuildingBlock specStartBlock = specBlocks.get(pair.getStart());
			BuildingBlock specEndBlock = specBlocks.get(pair.getEnd());

			BuildingBlock wcStartBlock = wrongContinuationBlocks.get(pair
					.getStart());
			BuildingBlock wcEndBlock = wrongContinuationBlocks.get(pair
					.getEnd());

			boolean startOk = true;
			if (wcStartBlock != null) {
				startOk = specStartBlock.containsBlock(wcStartBlock, false);
			}

			LOG.info("Specification start block contains wrong continuation start block ? "
					+ startOk);

			boolean endOk = true;
			if (wcEndBlock != null) {
				endOk = specEndBlock.containsBlock(wcEndBlock, true);
			}

			LOG.info("Specification end block contains wrong continuation end block ? "
					+ endOk);

			if (startOk && endOk) {
				for (LpoEdge edge : changedEdges) {
					edge.setFoldingEdge(false);
				}
			} else {
				return true;
			}
		}

		for (LpoEdge edge : changedRunEdges) {
			edge.setFoldingEdge(false);
		}

		return false;
	}

	private Vector<ActivityPair> findPossibleLabelPairs(
			Vector<Lpo> specification, WrongContinuation wrongContinuation) {

		Vector<Activity> specPreLabels = wrongContinuation
				.getSpecificationPreLabels();

		HashMap<Activity, Vector<Activity>> pairs = new HashMap<Activity, Vector<Activity>>();
		HashMap<Activity, HashMap<Activity, Vector<LpoEdge>>> edges = new HashMap<Activity, HashMap<Activity, Vector<LpoEdge>>>();

		Activity finalActivity = Activity.getActivity(Lpo.FINAL_LABEL);

		for (Activity activity : specPreLabels) {

			Vector<Activity> ends = new Vector<Activity>();
			pairs.put(activity, ends);

			edges.put(activity, new HashMap<Activity, Vector<LpoEdge>>());

			for (Vertex vertex : activity.getAllVertices()) {
				for (LpoEdge postEdge : vertex.getPostset()) {

					Activity end = postEdge.getEnd().getActivity();

					if (!end.equals(finalActivity)) {

						if (!ends.contains(end)) {
							ends.add(end);
							edges.get(activity).put(end, new Vector<LpoEdge>());
						}

						edges.get(activity).get(end).add(postEdge);
					}
				}
			}
		}

		Vector<ActivityPair> activityPairs = new Vector<ActivityPair>();
		for (Activity start : pairs.keySet()) {
			for (Activity end : pairs.get(start)) {
				ActivityPair pair = new ActivityPair(start, end, distance(
						start, end, specification));
				activityPairs.add(pair);

				pair.addAllPairEdges(edges.get(start).get(end));
			}
		}

		// sort pairs by distance
		System.err.println("before: " + activityPairs);
		Collections.sort(activityPairs, new ActivityPairDistanceComparator());
		System.err.println("after: " + activityPairs);

		System.err.println("Dist(x,y): "
				+ distance(Activity.getActivity("x"),
						Activity.getActivity("y"), specification));
		System.err.println("Dist(a,y): "
				+ distance(Activity.getActivity("a"),
						Activity.getActivity("y"), specification));

		LOG.info("Found " + activityPairs.size() + " possible activity pairs: "
				+ activityPairs);

		return activityPairs;
	}

	private int distance(Activity start, Activity end, Vector<Lpo> specification) {

		int min = Integer.MAX_VALUE;

		for (Lpo lpo : specification) {
			int dist = lpo.distance(start, end);

			if (dist < min) {
				min = dist;
			}
		}

		return min;
	}

	private HashMap<Activity, BuildingBlock> generateBlocksForWrongContinuation(
			Lpo run) {

		HashMap<Activity, BuildingBlock> wrongContinuationBlocks = new HashMap<Activity, BuildingBlock>();

		for (Vertex vertex : run.getAllVertices()) {
			foldingAlgo.addVertexToBlocks(wrongContinuationBlocks, vertex);
		}

		return wrongContinuationBlocks;
	}

}
