package de.fernunihagen.viptool.algorithm.folding.epcoflpo.logic;

import java.util.HashMap;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCFunction;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCNode;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCXorConnector;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.exec.EpcState;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Activity;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Lpo;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.LpoEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.Vertex;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.exec.LpoState;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.lpo.exec.WrongContinuation;

public class FindWrongContinuation {

	/**
	 * Logger for this class
	 */
	private static final Logger LOG = LoggerFactory
			.getLogger(FindWrongContinuation.class);

	private Vector<Lpo> specification;

	public FindWrongContinuation() {
		specification = new Vector<Lpo>();
	}

	public void addToSpecification(Vector<Lpo> lpos) {
		LOG.debug("Add to specification: " + lpos);

		specification.addAll(lpos);
	}

	public WrongContinuation runAlgo(EPC epc) {

		LOG.info("Searching wrong continuations...");

		epc.markWithInitialMarking();

		if (epc.getInitialNodes().size() > 1) {
			LOG.warn("Epc with multiple start nodes found!");
		}

		EPCNode startNode = epc.getInitialNodes().firstElement();

		Vector<LpoState> states = new Vector<LpoState>();
		for (Lpo lpo : specification) {
			// create LpoState for each lpo and add it to list
			states.add(new LpoState(lpo));
		}

		EpcState epcState = new EpcState(epc, 1, 1, states);

		// fire start node
		for (LpoState lpoState : states) {
			if (lpoState.canFire(startNode.getLabel())) {
				// fire start node
				Vector<Vertex> enabledVertices = lpoState
						.getEnabledVertices(Activity.getActivity(startNode
								.getLabel()));
				if (enabledVertices.size() > 1) {
					LOG.warn("LPO with multiple start activities found!");
				}
				lpoState.fire(enabledVertices.firstElement());
			} else {
				// can not fire start => wrong continuation
				Activity activity = Activity.getActivity(startNode.getLabel());
				Vertex wcVertex = epcState.getVertexForStep(0);
				HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>> specificationPresets = getSpecificationPresets(activity);
				return new WrongContinuation(epcState, (EPCFunction) startNode,
						wcVertex, specificationPresets,
						getSpecificationPreLabels(specificationPresets));
			}
		}

		LOG.debug("Searching wrong continuations: Lpos initialized. Start searching ...");

		WrongContinuation wc = findFirstWrongContiuation(epcState);

		if (wc != null) {
			LOG.info("Wrong continuation found: " + wc);
		} else {
			LOG.info("The epc " + epc
					+ " doesn't contain any extended behaviour.");
		}

		return wc;
	}

	private HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>> getSpecificationPresets(
			Activity activity) {

		LOG.info("Searching specification presets of label " + activity + "...");

		HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>> specificationPresets = new HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>>();

		for (Lpo lpo : specification) {

			Vector<Vertex> vertices = lpo.getVertices(activity.getLabel());
			if (vertices != null && !vertices.isEmpty()) {

				LOG.info("LPO " + lpo + " contains label " + activity);

				HashMap<Vertex, Vector<Vertex>> lpoMap = new HashMap<Vertex, Vector<Vertex>>();

				specificationPresets.put(lpo, lpoMap);

				for (Vertex vertex : vertices) {

					Vector<Vertex> preset = new Vector<Vertex>();
					lpoMap.put(vertex, preset);

					for (LpoEdge edge : vertex.getPreset()) {
						preset.add(edge.getStart());
					}

					LOG.info(lpo + ": Preset of " + activity + ": " + preset);
				}
			} else {
				LOG.info("LPO " + lpo + " doesn't contain label " + activity);
			}
		}

		return specificationPresets;
	}

	private Vector<Activity> getSpecificationPreLabels(
			HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>> specificationPresets) {

		LOG.info("Searching specification preset labels in "
				+ specificationPresets + " ...");

		Vector<Activity> preLabels = new Vector<Activity>();

		for (Lpo lpo : specificationPresets.keySet()) {
			for (Vector<Vertex> vertices : specificationPresets.get(lpo)
					.values()) {
				for (Vertex vertex : vertices) {
					if (!preLabels.contains(vertex.getActivity())) {
						LOG.debug(lpo + ": Preset " + vertices + ": New label "
								+ vertex.getActivity() + " found.");
						preLabels.add(vertex.getActivity());
					} else {
						LOG.debug(lpo + ": Preset " + vertices + ": Label "
								+ vertex.getActivity() + " was already found.");
					}
				}
			}
		}

		return preLabels;
	}

	private WrongContinuation findFirstWrongContiuation(EpcState state) {

		LOG.info("Recursive search: " + state + ", " + state.getEnabledLpos());

		WrongContinuation wc = runWithoutDecission(state);
		if (wc != null) {
			return wc;
		}

		LOG.debug("Need alternatives: " + state + ", " + state.getEnabledLpos());

		for (EPCXorConnector alt : state.getActiveAlternatives()) {
			for (EPCEdge e : alt.getPostset()) {
				EpcState es = state.clone(alt.getPostset().indexOf(e) + 1, alt
						.getPostset().size());

				EPCXorConnector alternative = (EPCXorConnector) es.getEpc()
						.translateToClonedNode(alt);

				EPCNode choosenNode = es.getEpc().translateToClonedNode(
						e.getEnd());

				LOG.info(state + ": choosen alternative: " + choosenNode
						+ " of " + alternative);

				try {
					es.fire(alternative, choosenNode);
				} catch (Exception ex) {
					LOG.error("Can not fire node " + alternative
							+ "! (node.canFire: " + alternative.canFire() + ")");
					ex.printStackTrace();
				}

				wc = findFirstWrongContiuation(es);

				if (wc != null) {
					return wc;
				}
			}
		}

		LOG.info(state.shortName() + ": alternative finished without errors.");
		return null;
	}

	private WrongContinuation runWithoutDecission(EpcState state) {

		while (!state.getActiveConnectors().isEmpty()
				|| !state.getActiveFunctions().isEmpty()) {
			fireConnectors(state);
			WrongContinuation wc = fireFunctions(state);
			if (wc != null) {
				return wc;
			}
		}

		return null;
	}

	private void fireConnectors(EpcState epcState) {
		LOG.debug(epcState.shortName()
				+ ": Firing all available connectors without decission...");

		while (!epcState.getActiveConnectors().isEmpty()) {

			LOG.debug(epcState.shortName() + ": Firing "
					+ epcState.getActiveConnectors().size()
					+ " available connectors...");
			while (!epcState.getActiveConnectors().isEmpty()) {
				EPCConnector conn = epcState.getActiveConnectors()
						.firstElement();

				try {
					epcState.fire(conn, null);
					LOG.debug("fireConnectors: New epc state: " + epcState);
				} catch (Exception e) {
					LOG.error("Can not fire connector " + conn
							+ "! (node.canFire: " + conn.canFire() + ")");
					LOG.error("EPC of connector: " + conn.getEpc());
					LOG.error("EPC of state: " + epcState.getEpc());
					LOG.error("EPC state: " + epcState);

					String desc = "{";
					for (EPCEdge edge : conn.getPreset()) {
						desc += "\"" + edge.getStart() + "\" ";
					}
					desc += "} -- " + conn + " -- {";
					for (EPCEdge edge : conn.getPostset()) {
						desc += "\"" + edge.getEnd() + "\" ";
					}
					desc += "}";
					LOG.error("Description of " + conn + ": " + desc);

					e.printStackTrace();
					System.exit(0);
				}
			}
		}
	}

	private WrongContinuation fireFunctions(EpcState state) {

		LOG.info("Firing all available functions of " + state);

		Vector<EPCFunction> functions = new Vector<EPCFunction>();
		functions.addAll(state.getActiveFunctions());
		for (EPCFunction node : functions) {
			if (node instanceof EPCFunction) {
				EPCFunction function = (EPCFunction) node;
				Activity activity = Activity.getActivity(function.getLabel());

				LOG.info(state.shortName() + ": Fire " + activity + ", "
						+ state.getEnabledLpos().size() + " enabled lpos");

				try {
					state.fire(node, null);
				} catch (Exception e) {
					LOG.error("Can not fire node " + node + "! (node.canFire: "
							+ node.canFire() + ")");
					e.printStackTrace();
				}

				LOG.debug(state.shortName() + ": firing " + function);

				Vector<LpoState> possibleLpos = new Vector<LpoState>();
				for (LpoState lpo : state.getEnabledLpos()) {
					if (lpo.canFire(activity)) {
						Vertex vertex = lpo.getEnabledVertices(activity)
								.firstElement();
						lpo.fire(vertex);
						possibleLpos.add(lpo);
					} else {
						LOG.info(state.shortName() + ": LPO " + lpo
								+ " can't fire.");
					}
				}

				if (possibleLpos.isEmpty()) {
					LOG.info(state.shortName() + ": Wrong continuation found!");

					Vertex wcVertex = state.getVertexForStep(state
							.getFireSequence().size() - 1);
					HashMap<Lpo, HashMap<Vertex, Vector<Vertex>>> specificationPresets = getSpecificationPresets(activity);

					return new WrongContinuation(state, function, wcVertex,
							specificationPresets,
							getSpecificationPreLabels(specificationPresets));
				} else {
					LOG.debug(state.shortName() + ": enabled lpos: "
							+ possibleLpos);
					state.setEnabledLpos(possibleLpos);
				}
			}
		}

		return null;
	}

}
