package de.fernunihagen.viptool.test.algorithm.folding.epcoflpo.datastructure.epc;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Vector;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPC;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCEdge;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCFunction;
import de.fernunihagen.viptool.algorithm.folding.epcoflpo.datastructure.epc.EPCXorConnector;

public class TestEPCFunction {

	private EPC epc;
	private EPCXorConnector preXor;
	private EPCXorConnector postXor;

	@Before
	public void setUp() throws Exception {
		epc = new EPC();

		preXor = epc.addXorConnector();
		postXor = epc.addXorConnector();
	}

	@After
	public void tearDown() throws Exception {
	}

	@Test
	public void testCanFire() {
		EPCFunction function = epc.addFunction("testFunction");

		epc.connect(preXor, function);
		epc.connect(function, postXor);

		// TODO: test mark sources

		try {
			function.getPreset().firstElement().setMarked(false, null);
		} catch (Exception e) {
			e.printStackTrace();
		}
		assertFalse(function.canFire());

		try {
			Vector<EPCFunction> markSources = new Vector<EPCFunction>();
			markSources.add(epc.addFunction("TestFunction"));
			function.getPreset().firstElement().setMarked(true, markSources);
		} catch (Exception e) {
			e.printStackTrace();
		}
		assertTrue(function.canFire());
	}

	@Test
	public void testFire() {
		EPCFunction function = epc.addFunction("testFunction");

		epc.connect(preXor, function);
		epc.connect(function, postXor);

		try {
			Vector<EPCFunction> markSources = new Vector<EPCFunction>();
			markSources.add(epc.addFunction("TestFunction"));
			function.getPreset().firstElement().setMarked(true, markSources);
		} catch (Exception e) {
			e.printStackTrace();
		}

		try {
			function.fire(null);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		assertFalse(function.getPreset().firstElement().isMarked());
		assertTrue(function.getPostset().firstElement().isMarked());
	}

	@Test
	public void testGetLabel() {
		EPCFunction function = epc.addFunction("testFunction");
		assertTrue(function.getLabel().equals("testFunction"));
	}

	@Test
	public void testAddIncomingEdge() {
		EPCFunction function = epc.addFunction("testFunction");
		EPCEdge edge = EPCEdge.getEdge(preXor, function, epc);
		function.addIncomingEdge(edge);

		assertTrue(function.getPreset().size() == 1);
		assertTrue(function.getPreset().contains(edge));
	}

	@Test
	public void testRemoveIncomingEdge() {
		EPCFunction function = epc.addFunction("testFunction");
		EPCEdge edge = EPCEdge.getEdge(preXor, function, epc);
		function.addIncomingEdge(edge);

		assertTrue(function.getPreset().contains(edge));

		function.removeIncomingEdge(edge);

		assertTrue(function.getPreset().isEmpty());
	}

	@Test
	public void testAddOutgoingEdge() {
		EPCFunction function = epc.addFunction("testFunction");
		EPCEdge edge = EPCEdge.getEdge(function, postXor, epc);
		function.addOutgoingEdge(edge);

		assertTrue(function.getPostset().size() == 1);
		assertTrue(function.getPostset().contains(edge));
	}

	@Test
	public void testRemoveOutgoingEdge() {
		EPCFunction function = epc.addFunction("testFunction");
		EPCEdge edge = EPCEdge.getEdge(function, postXor, epc);
		function.addOutgoingEdge(edge);

		assertTrue(function.getPostset().contains(edge));

		function.removeOutgoingEdge(edge);

		assertTrue(function.getPostset().isEmpty());
	}

	@Test
	public void testGetPreset() {
		EPCFunction function = epc.addFunction("testFunction");
		assertTrue(function.getPreset().isEmpty());
	}

	@Test
	public void testGetPostset() {
		EPCFunction function = epc.addFunction("testFunction");
		assertTrue(function.getPostset().isEmpty());
	}

	@Test
	public void testConnectToPredecessor() {
		EPCFunction function = epc.addFunction("testFunction");

		function.connectToPredecessor(preXor);

		assertTrue(function.getPreset().size() == 1);
		assertTrue(function.getPreset().firstElement().getStart()
				.equals(preXor));

		boolean found = false;
		for (EPCEdge edge : preXor.getPostset()) {
			if (edge.getEnd().equals(function)) {
				found = true;
			}
		}
		assertTrue(found);
	}

	@Test
	public void testConnectToSuccessor() {
		EPCFunction function = epc.addFunction("testFunction");

		function.connectToSuccessor(postXor);

		assertTrue(function.getPostset().size() == 1);
		assertTrue(function.getPostset().firstElement().getEnd()
				.equals(postXor));

		boolean found = false;
		for (EPCEdge edge : postXor.getPreset()) {
			if (edge.getStart().equals(function)) {
				found = true;
			}
		}
		assertTrue(found);
	}

	@Test
	public void testDisconnectFromPredecessor() {
		EPCFunction function = epc.addFunction("testFunction");

		function.connectToPredecessor(preXor);

		assertTrue(function.getPreset().size() == 1);
		assertTrue(function.getPreset().firstElement().getStart()
				.equals(preXor));

		function.disconnectFromPredecessor(preXor);

		assertTrue(function.getPreset().isEmpty());

		for (EPCEdge edge : preXor.getPostset()) {
			if (edge.getEnd().equals(function)) {
				fail("Predecessor still connected!");
			}
		}
	}

	@Test
	public void testDisconnectFromSuccessor() {
		EPCFunction function = epc.addFunction("testFunction");

		function.connectToSuccessor(postXor);

		assertTrue(function.getPostset().size() == 1);
		assertTrue(function.getPostset().firstElement().getEnd()
				.equals(postXor));

		function.disconnectFromSuccessor(postXor);

		assertTrue(function.getPostset().isEmpty());

		for (EPCEdge edge : postXor.getPreset()) {
			if (edge.getStart().equals(function)) {
				fail("Successor still connected!");
			}
		}
	}

	@Test
	public void testGetEpc() {
		EPCFunction function = epc.addFunction("testFunction");
		assertTrue(function.getEpc().equals(epc));
	}

}
